from __future__ import annotations

from zoneinfo import ZoneInfo

HONG_KONG = ZoneInfo("Asia/Hong_Kong")
TOKYO = ZoneInfo("Asia/Tokyo")
US_CENTRAL = ZoneInfo("US/Central")
US_EASTERN = ZoneInfo("US/Eastern")


__all__ = ["HONG_KONG", "TOKYO", "US_CENTRAL", "US_EASTERN"]
